<?php

/**
 * License library for op5 products
 */
class op5lic {
	/*
	 * fetch_metric returns this for infinit value. Always use === or !== to
	 * compare
	 */
	const METRIC_INFINITY = PHP_INT_MAX;
	const DEFAULT_LICENSE_FILE = '/etc/op5license/op5license.xml';
	const NINJA_PATH = '/opt/monitor/op5/ninja/';

	const OP5_LICENSE_URL = 'http://www.op5.com/explore-op5-monitor/price-and-comparison/';

	/**
	 *
	 * @var array
	 */
	protected $metrics = array ();
	/**
	 * Contains an array of attributes set depending on license type.
	 *
	 * The attributes available is specified in this array from the beginning.
	 * No attributes should be added or deleted when reading the license.
	 *
	 * Attributes may be added in later versions of this library though.
	 *
	 * @var array
	 */
	/* Start with infromation representing an unlicensed installation */
	private $properties = array ('always_show_licenseinfo' => true,
		'trial' => false,'unlicensed' => true,'unlimited' => false);
	/**
	 *
	 * @var array
	 */
	private $customer = array ();
	/**
	 *
	 * @var int
	 */
	private $trial_issued = 0;
	/**
	 *
	 * @var int
	 */
	private $trial_expires = 0;

	/**
	 *
	 * @var unknown
	 */
	private $load_messages = array ();

	/**
	 * Constructor
	 *
	 * @param $file string
	 *        	optional parameter. Filename of licensefile
	 * @throws op5licenseException
	 * @throws LicenseFileException
	 * @return void
	 */
	public function __construct($file = self::DEFAULT_LICENSE_FILE) {
		if (!file_exists($file)) {
			$this->load_trial_license();
		} else if (!$this->verify_signature($file)) {
			// FIXME: Invalid signature, manipulated?
			$this->load_messages[] = "Invalid license signature";
		} else {
			$license_data = @simplexml_load_file($file);

			/*
			 * This case should not really be possible to achive, not even with
			 * a test, but if the case is that the verifier can parse the
			 * license, validate the signature, but simplexml can't parse, this
			 * is a valid result. But since it shouldn't be possible to get
			 * here, we can't test it, so keep it out of code coverage analysis
			 */
			// @codeCoverageIgnoreStart
			if (!$license_data) {
				// FIXME: Invalid xml
				$this->load_messages[] = "Could not parse license XML";
			}
			// @codeCoverageIgnoreEnd

			/* Read and store license data */
			if (!$this->analyze_license($license_data)) {
				$this->load_messages[] = "Error reading license content";
			}
		}
	}

	/**
	 * Lookup if a license type provieds a certain functionality.
	 *
	 * To see a list of which attributes is allowed, look at the $properties
	 * variable above.
	 *
	 * If asking for an undefined attribute, assume false, to reduce risk for
	 * conflicting version between license lib and product.
	 *
	 * @param $name string
	 */
	public function fetch_property($prop) {
		if (isset($this->properties[$prop])) {
			return $this->properties[$prop];
		}
		return false;
	}

	/**
	 * Queries the license if the specified usage is ok according to the
	 * license.
	 *
	 * This method does not fetch any information about the system, so
	 * everything needs to be provided through the metric-array. (which one
	 * excpetion, se below)
	 *
	 * The metrics array should contain a list of metrics, and their usage, for
	 * example, a system with 47 monitor hosts, and two pollers should run:
	 * query(array('monitor' => 47, 'pollers' => 2)).
	 *
	 * The exception is license expiration time, which is the metric
	 * 'time_expire' as unix timestamp. If not specified, the query-method fills
	 * in that field before running the test with the system time. To explicitly
	 * skip expiration time check, specify the metric 'time_expire' => false
	 *
	 * @param $metrics array
	 *        	Metrics to tests
	 * @param $message array
	 *        	reference variable that will be set to an array of status
	 *        	messages. Optional parameter.
	 * @return boolean
	 */
	public function query(array $metrics, &$message = false) {
		$message = $this->load_messages;
		$valid = true;

		/* Handle all keys as lowercase, to make it insensitive */
		$metrics = array_change_key_case($metrics, CASE_LOWER);

		if (!array_key_exists('time_expire', $metrics)) {
			/* If time isn't specified, assume time is right now */
			$metrics['time_expire'] = time();
		} else {
			/* Make it possible to explicitly ignore time */
			if ($metrics['time_expire'] === false) {
				unset($metrics['time_expire']);
			}
		}

		foreach ($metrics as $metric => $usage) {
			$license_size = $this->fetch_metric($metric);

			if ($license_size === self::METRIC_INFINITY) {
				/*
				 * License size of self::METRIC_INFINITY means infinity, always
				 * valid, therefore don't do anything here... Just keep this
				 * case for code coverage analysis in unit tests
				 */
			} else {
				/*
				 * License size isn't op5lic::METRIC_INFINITY, treat it is a
				 * normal license
				 */
				if ($license_size === false)
					$license_size = 0;

				if ($usage > $license_size) {
					$valid = false;
					if ($metric == 'time_expire') {
						$message[] = "Your license has expired " . $this->format_time_diff(
							$usage - $license_size) . " ago";
					} else {
						$message[] = "Your usage for $metric is $usage, your license allows $license_size";
					}
				}
			}
		}

		return $valid;
	}

	/**
	 * Takes a metric, a usage amount, and a threshold as a factor of the
	 * license size.
	 *
	 * Return true if theshold isn't reached.
	 *
	 * Example:
	 * $threshold_reached = query_threshold( 'monitor', 12, 0.9 );
	 *
	 * @param
	 *        	$metric
	 * @param
	 *        	$usage
	 * @param
	 *        	$threshold
	 * @param $message array
	 *        	containing the error messages
	 * @return boolean
	 */
	public function query_threshold($metric, $usage, $threshold,
		&$message = false) {
		$message = array ();
		$licensed = $this->fetch_metric($metric);
		if ($licensed)
			$fraction = (float) $usage / (float) $licensed;
		else
			$fraction = 0;

		if ($licensed === self::METRIC_INFINITY) {
			$message[] = sprintf("No limitation in license size for %s",
				$metric);
			return true;
		}
		if ($licensed == 0 && $usage > 0) {
			$message[] = sprintf(
				"You are not licensed to have any %s, you use %d", $metric,
				$usage);
		} else if ($usage > $licensed) {
			$message[] = sprintf(
				"You are exceeding your license (you use %d, you may use %d) for %s",
				$usage, $licensed, $metric);
		} else if ($fraction >= $threshold) {
			$message[] = sprintf(
				"You are using %0.1f%% (%d out of %d) of your license for %s",
				$fraction * 100, $usage, $licensed, $metric);
		}
		if ($message)
			return false;
		$message[] = sprintf("You are using %0.1f%% of your license for %s",
			$fraction * 100, $metric);
		return true;
	}

	/**
	 * Given a metric, this method returns the number of objects allowed for
	 * that specific metric.
	 *
	 * A metric is something measurable in the system. Some are:
	 * - 'monitor' => number of hosts allowed for monitor
	 * - 'logserver' => number of hosts allowed for logserver
	 * - 'pollers' => how many pollers the system is allowed to handle
	 * - and many more, which is up to the license file to define.
	 *
	 * If the license doesn't define the metric, it is assumed that no objects
	 * is allowed, which means 0.
	 *
	 * If the license allowes infinit number of objects, this metric returns the
	 * magic value of self::METRIC_INFINITY. Make sure you use strict comparison
	 * of the that value, i.e. === or !==
	 *
	 * @param $product integer
	 *        	number of objects allowed for the given metric, or
	 *        	self::METRIC_INFINITY
	 */
	public function fetch_metric($metric) {
		if (!isset($this->metrics[$metric])) {
			/*
			 * Trial only sets some metadata, which shouldn't be overridden,
			 * like expiration. If a metric isn't defined in the license, we
			 * should allow everything
			 */
			if ($this->fetch_property('unlimited')) {
				return self::METRIC_INFINITY;
			}

			/*
			 * If a metric isn't defined, and license isn't trial, it means no
			 * amount of the metric is allowed
			 */
			return 0;
		}
		return $this->metrics[$metric];
	}

	/**
	 * Fetch a list of defined metrics for the license.
	 * Notable is that the license might grant access to other metrics than what
	 * is returned here, for example a trial allows full access to everything.
	 * This metod should only be used to visualize the status of the license
	 * when displaying lists of metrics, not for any license checking.
	 *
	 * For license checking, use query() instead. For statistics given a
	 * specific metric, the fetch_metric() is often better to use.
	 */
	public function fetch_license_metrics() {
		return $this->metrics;
	}

	/**
	 * Returns information about the customer
	 */
	public function fetch_customer_info() {
		return $this->customer;
	}

	/**
	 * Returns a list of errors produced when reading the license.
	 * Those are always included in the message array when doing a query
	 */
	public function fetch_load_messages() {
		return $this->load_messages;
	}

	/**
	 * "Old" API.
	 * deprecated.
	 */
	public function fetch_load_errors() {
		return $this->fetch_load_messages();
	}

	/**
	 * Verify license file signature
	 *
	 * @return boolean
	 *
	 */
	private function verify_signature($filename) {
		exec("/opt/op5sys/bin/op5license-verify " . $filename, $verify_output,
			$verify_return);
		return $verify_return === 0;
	}

	/**
	 *
	 * @param $license SimpleXMLElement
	 */
	private function analyze_license($license) {
		$verifier = (string) $license['verifier'];
		switch ($verifier) {
		case 'op5v1':
			return $this->analyze_license_op5v1($license);
		case 'op5v2':
			return $this->analyze_license_op5v2($license);
		}
		// FIXME: Invalid verifier
		return false;
	}

	/**
	 *
	 * @param $licence SimpleXMLElement
	 *        	license file parsed
	 */
	private function analyze_license_op5v2(SimpleXMLElement $license) {
		$allowed_types = array ('regular' => true,'site' => true,
			'evaluation' => true,'poc' => true,'enterprise' => true,
			'basic' => true,'free' => true);

		/* Initialize with default values */
		$this->metrics['time_issue'] = 0;
		$this->metrics['time_expire'] = self::METRIC_INFINITY;

		$this->properties['always_show_licenseinfo'] = false;
		$this->properties['unlicensed'] = false;
		$this->properties['unlimited'] = false;
		$this->properties['trial'] = false;

		/* Extract item tags */
		foreach ($license->item as $it) {
			$name = (string) $it['name'];
			$value = (string) $it['value'];

			/* Normal fields */
			switch (strtolower($name)) {
			case 'issuedtocustomernumber':
				$this->customer['number'] = $value;
				break;
			case 'issuedtoname':
				$this->customer['name'] = $value;
				break;
			case 'issuedate':
				$this->metrics['time_issue'] = strtotime($value);
				if ($this->metrics['time_issue'] === false)
					$this->metrics['time_issue'] = 0;
				break;
			case 'expiredate':
				$this->metrics['time_expire'] = strtotime($value);
				if ($this->metrics['time_expire'] === false)
					$this->metrics['time_expire'] = self::METRIC_INFINITY;
				break;
			}
		}

		/* Extract products */
		if ($license->products) {
			foreach ($license->products->children() as $product) {
				$prodname = (string) $product->getName();
				foreach ($product->item as $it) {
					$name = (string) $it['name'];
					$value = (string) $it['value'];
					if (preg_match('/^(.*)LicenseType$/', $name, $matches)) {
						/* Name ends with LicenseType */
						$type = strtolower($value);
						if (!isset($allowed_types[$type])) {
							/*
							 * TODO: Should probably be some error message
							 * here...
							 */
							$this->load_unlicensed();
							return false;
						}
						if ($type == 'site') {
							/* License Type: SITE */
							$this->metrics[strtolower($matches[1])] = op5lic::METRIC_INFINITY;
						}
					} else if (preg_match('/^(.*)Hosts$/', $name, $matches)) {
						/* Name ends with Hosts, which means metric */
						if (!isset($this->metrics[strtolower($matches[1])]))
							$this->metrics[strtolower($matches[1])] = intval(
								$value);
					}
				}

				/* Fetch extensions */
				if ($product->extensions) {
					foreach ($product->extensions->item as $ext) {
						$extname = (string) $ext['name'];
						$extvalue = (string) $ext['value'];
						switch (strtolower($extvalue)) {
						case 'no':
							$extvalueint = 0;
							break;
						case 'yes':
							if ((string) $ext['Hosts']) {
								$extvalueint = (int) (string) $ext['Hosts'];
							} else {
								$extvalueint = 1;
							}
							break;
						default:
							$extvalueint = intval($extvalue);
						}
						/*
						 * Never overwrite. For example, the "LogserverHosts"
						 * has higher precedence than extension "logserver"
						 */
						if (!isset($this->metrics[strtolower($extname)]))
							$this->metrics[strtolower($extname)] = $extvalueint;
					}
				}
			}
		}
		return true;
	}

	/**
	 *
	 * @param $licence SimpleXMLElement
	 *        	license file parsed
	 */
	private function analyze_license_op5v1(SimpleXMLElement $license) {
		$allowed_types = array ('regular' => true,'site' => true,
			'evaluation' => true,'poc' => true,'enterprise' => true,
			'basic' => true,'free' => true);

		/* Initialize with default values */
		$this->metrics['time_issue'] = 0;
		$this->metrics['time_expire'] = self::METRIC_INFINITY;

		$this->properties['always_show_licenseinfo'] = false;
		$this->properties['unlicensed'] = false;
		$this->properties['unlimited'] = false;
		$this->properties['trial'] = false;

		/* Extract item tags */
		foreach ($license->item as $it) {
			$name = (string) $it['name'];
			$value = (string) $it['value'];

			if (preg_match('/^(.*)LicenseType$/', $name, $matches)) {
				/* Special case if name ends with LicenseType */
				$type = strtolower($value);
				if (!isset($allowed_types[$type])) {
					/* TODO: Should probably be some error message here... */
					$this->load_unlicensed();
					return false;
				}
				if ($type == 'site') {
					/* License Type: SITE */
					$this->metrics[strtolower($matches[1])] = op5lic::METRIC_INFINITY;
				}
			} else if (preg_match('/^(.*)Hosts$/', $name, $matches)) {
				/* Special case if name ends with Hosts, which means metric */
				if (!isset($this->metrics[strtolower($matches[1])]))
					$this->metrics[strtolower($matches[1])] = intval($value);
			} else {
				/* Normal fields */
				switch (strtolower($name)) {
				case 'issuedtocustomernumber':
					$this->customer['number'] = $value;
					break;
				case 'issuedtoname':
					$this->customer['name'] = $value;
					break;
				case 'issuedtoemail':
					$this->customer['email'] = $value;
					break;
				case 'issuedate':
					$this->metrics['time_issue'] = strtotime($value);
					if ($this->metrics['time_issue'] === false)
						$this->metrics['time_issue'] = 0;
					break;
				}
			}
		}

		/* Fetch extensions */
		if ($license->extension) {
			foreach ($license->extension->item as $ext) {
				$extname = (string) $ext['name'];
				$extvalue = (string) $ext['value'];
				switch (strtolower($extvalue)) {
				case 'no':
					$extvalueint = 0;
					break;
				case 'yes':
					$extvalueint = 1;
					break;
				default:
					$extvalueint = intval($extvalue);
				}
				/*
				 * Never overwrite. For example, the "LogserverHosts" has higher
				 * precedence than extension "logserver"
				 */
				if (!isset($this->metrics[strtolower($extname)]))
					$this->metrics[strtolower($extname)] = $extvalueint;
			}
		}

		return true;
	}

	/**
	 * Invent some values for the trial license
	 */
	private function load_trial_license() {
		exec('php ' . self::NINJA_PATH . '/index.php cli/license_start',
			$output, $exit_code);
		$output = intval(trim(implode("", $output)));
		if ($exit_code || !$output) {
			$output = time(); // If nothing found, assume the license starts now
		}

		$start_time = intval($output);
		$end_time = $start_time + 3600 * 24 * 30; // Add 30 days in seconds

		$this->metrics['time_issue'] = $start_time;
		$this->metrics['time_expire'] = $end_time;

		if (time() <= $end_time) {
			/* License Type: TRIAL */
			$this->properties['always_show_licenseinfo'] = true;
			$this->properties['unlicensed'] = false;
			$this->properties['unlimited'] = true;
			$this->properties['trial'] = true;

			$this->load_messages[] = '<a href="'.self::OP5_LICENSE_URL.'">You are using a trial license, contact op5 for more information</a>';
			$this->load_messages[] = 'Your license will expire in ' .
				 $this->format_time_diff($end_time - time());
		} else {
			/* License Type: UNLICENSED */
			$this->properties['always_show_licenseinfo'] = true;
			$this->properties['unlicensed'] = true;
			$this->properties['unlimited'] = false;
			$this->properties['trial'] = false;

			$this->load_messages[] = '<a href="'.self::OP5_LICENSE_URL.'">Your trial license has expired, contact op5 for more information</a>';
		}
	}

	/**
	 * Reset information to unlicensed
	 */
	private function load_unlicensed() {
		$this->metrics = array ();

		/* License Type: UNLICENSED */
		$this->properties['always_show_licenseinfo'] = true;
		$this->properties['unlicensed'] = true;
		$this->properties['unlimited'] = false;
		$this->properties['trial'] = false;
	}

	/**
	 * Format time difference
	 *
	 * @param $deltat int
	 *        	number of seconds
	 * @return a string representing the interval as X days Y hours Z minutes
	 */
	private function format_time_diff($deltat) {
		$parts = array ();
		$parts['day'] = intval($deltat / (60 * 60 * 24));
		$parts['hour'] = intval($deltat / (60 * 60)) % 24;
		$parts['minute'] = intval($deltat / 60) % 60;
		$parts['second'] = intval($deltat) % 60;

		$precision = 0;
		$ret = array ();
		foreach ($parts as $k => $v) {
			/* If we found a suffix, increment */
			if ($precision)
				$precision++;

				/* If the suffix has a value, possibly print it */
			if ($v > 0) {
				/* If we havn't already found a suffix, we have now */
				if (!$precision)
					$precision = 1;

					/* Only display a precision of two suffixes */
				if ($precision <= 2)
					$ret[] = $v . ' ' . $k . ($v > 1 ? 's' : '');
			}
		}

		return join(' ', $ret);
	}
}
